"""
Routines to manage the setup and invocation of pyinstaller
"""
# Import python libs
import os
import subprocess
import shutil


SPEC = """# -*- mode: python ; coding: utf-8 -*-

block_cipher = None


a = Analysis([r"{s_path}"],
             pathex=[r"{cwd}"],
             binaries={binaries},
             datas={datas},
             hiddenimports={imports},
             hookspath=[],
             runtime_hooks=[],
             excludes=[],
             win_no_prefer_redirects=False,
             win_private_assemblies=False,
             cipher=block_cipher,
             noarchive=False)
pyz = PYZ(a.pure, a.zipped_data,
             cipher=block_cipher)
"""

SPEC_EXE = """

exe = EXE(pyz,
          a.scripts,
          a.binaries,
          a.zipfiles,
          a.datas,
          [],
          name=r"{name}",
          exclude_binaries={onedir},
          debug=False,
          bootloader_ignore_signals=False,
          strip=False,
          upx=True,
          upx_exclude=[],
          runtime_tmpdir={runtime_tmpdir},
          console=True )
"""

SPEC_COLL = """

exe = EXE(pyz,
          a.scripts,
          [],
          name=r"{name}",
          exclude_binaries={onedir},
          debug=False,
          bootloader_ignore_signals=False,
          strip=False,
          upx=True,
          upx_exclude=[],
          runtime_tmpdir={runtime_tmpdir},
          console=True )

coll = COLLECT(exe,
               a.binaries,
               a.zipfiles,
               a.datas,
               strip=False,
               upx=True,
               upx_exclude=[],
               name=r"{name}")
"""


def mk_spec(hub, bname):
    """
    Create a spec file to build from
    """
    opts = hub.pop_build.BUILDS[bname]
    datas = []
    imps = []
    kwargs = {
        "s_path": opts["s_path"],
        "cwd": os.getcwd(),
        "name": opts["name"],
    }
    for imp in opts["imports"]:
        imp = imp.replace("\\", "\\\\")
        imps.append(imp)
    for data in opts["datas"]:
        src, dst = data.split(os.pathsep)
        src = src.replace("\\", "\\\\")
        dst = dst.replace("\\", "\\\\")
        datas.append((src, dst))
    kwargs["datas"] = datas.__repr__()
    kwargs["imports"] = imps.__repr__()
    kwargs["binaries"] = opts.get("binaries", []).__repr__()
    kwargs["runtime_tmpdir"] = opts.get("pyinstaller_runtime_tmpdir", None)
    if isinstance(kwargs["runtime_tmpdir"], str):
        kwargs["runtime_tmpdir"] = kwargs["runtime_tmpdir"].__repr__()
    if "--onedir" in hub.pop_build.BUILDS[bname]["pypi_args"]:
        kwargs["onedir"] = True
        kwargs["name"] = "run"
        spec = SPEC.format(**kwargs)
        spec = spec + SPEC_COLL.format(**kwargs)
    else:
        kwargs["onedir"] = False
        spec = SPEC.format(**kwargs)
        spec = spec + SPEC_EXE.format(**kwargs)
    with open(opts["spec"], "w+") as wfh:
        wfh.write(spec)
    opts["cmd"] += f' {opts["spec"]}'


def call(hub, bname):
    opts = hub.pop_build.BUILDS[bname]
    dname = os.path.dirname(opts["s_path"])
    if not os.path.isdir(dname):
        os.makedirs(os.path.dirname(opts["s_path"]))
    shutil.copy(opts["run"], opts["s_path"])
    subprocess.call(opts["cmd"], shell=True)



# Import python libs
import os
import uuid
import subprocess
import shutil
import tempfile


def __init__(hub):
    hub.pop.sub.load_subdirs(hub.pop_build)
    hub.pop_build.BUILDS = {}
    hub.pop_build.SYSTEMD = ("rhel7", "rhel8", "arch", "debian10")
    hub.pop_build.SYSTEMD_DIR = "usr/lib/systemd/system/"
    hub.pop_build.SYSV = ("rhel5", "rhel6")
    hub.pop_build.SYSV_DIR = "etc/init.d"


def cli(hub):
    """
    Execute the routine from the CLI
    """
    hub.pop.config.load(["pop_build"], cli="pop_build")
    hub.pop_build.init.builder(
        hub.OPT["pop_build"]["name"],
        hub.OPT["pop_build"]["requirements"],
        hub.OPT["pop_build"]["system_site"],
        hub.OPT["pop_build"]["exclude"],
        hub.OPT["pop_build"]["directory"],
        hub.OPT["pop_build"]["dev_pyinstaller"],
        hub.OPT["pop_build"]["pyinstaller_runtime_tmpdir"],
        hub.OPT["pop_build"]["build"],
        hub.OPT["pop_build"]["pkg"],
        hub.OPT["pop_build"]["onedir"],
        hub.OPT["pop_build"]["pyenv"],
        hub.OPT["pop_build"]["run"],
        hub.OPT["pop_build"]["no_clean"],
        hub.OPT["pop_build"]["locale_utf8"],
        hub.OPT["pop_build"]["pkg_tgt"],
        hub.OPT["pop_build"]["pkg_builder"],
        hub.OPT["pop_build"]["srcdir"],
    )


def new(
    hub,
    name,
    requirements,
    sys_site,
    exclude,
    directory,
    dev_pyinst=False,
    pyinstaller_runtime_tmpdir=None,
    build=None,
    pkg=None,
    onedir=False,
    pyenv="system",
    run="run.py",
    locale_utf8=False,
    pkg_tgt=None,
    pkg_builder=None,
    srcdir=None,
):
    venv_dir = tempfile.mkdtemp()
    is_win = os.name == "nt"
    if is_win:
        python_bin = os.path.join(venv_dir, "Scripts", "python")
        s_path = os.path.join(venv_dir, "Scripts", name)
    else:
        python_bin = os.path.join(venv_dir, "bin", "python3")
        if locale_utf8:
            s_path = "env PYTHONUTF8=1 LANG=POSIX " + os.path.join(
                venv_dir, "bin", name
            )
        else:
            s_path = os.path.join(venv_dir, "bin", name)
    bname = str(uuid.uuid1())
    requirements = os.path.join(directory, requirements)
    hub.pop_build.BUILDS[bname] = {
        "name": name,
        "build": build,
        "pkg": pkg,
        "pkg_tgt": pkg_tgt,
        "pkg_builder": pkg_builder,
        "binaries": [],
        "is_win": is_win,
        "exclude": exclude,
        "requirements": requirements,
        "sys_site": sys_site,
        "dir": os.path.abspath(directory),
        "srcdir": srcdir,
        "dev_pyinst": dev_pyinst,
        "pyinstaller_runtime_tmpdir": pyinstaller_runtime_tmpdir,
        "run": os.path.join(directory, run),
        "spec": os.path.join(directory, f"{name}.spec"),
        "pybin": python_bin,
        "s_path": s_path,
        "venv_dir": venv_dir,
        "vroot": os.path.join(venv_dir, "lib"),
        "onedir": onedir,
        "all_paths": set(),
        "imports": set(),
        "datas": set(),
        "cmd": f"{python_bin} -B -OO -m PyInstaller ",
        "pyenv": pyenv,
        "pypi_args": [
            s_path,
            "--log-level=INFO",
            "--noconfirm",
            "--onedir" if onedir else "--onefile",
            "--clean",
        ],
        "locale_utf8": locale_utf8,
    }
    req = hub.pop_build.init.mk_requirements(bname)
    hub.pop_build.BUILDS[bname]["req"] = req
    return bname


def mk_requirements(hub, bname):
    opts = hub.pop_build.BUILDS[bname]
    req = os.path.join(opts["dir"], "__build_requirements.txt")
    with open(opts["requirements"], "r") as rfh:
        data = rfh.read()
    with open(req, "w+") as wfh:
        wfh.write(data)
    return req


def builder(
    hub,
    name,
    requirements,
    sys_site,
    exclude,
    directory,
    dev_pyinst=False,
    pyinstaller_runtime_tmpdir=None,
    build=None,
    pkg=None,
    onedir=False,
    pyenv="system",
    run="run.py",
    no_clean=False,
    locale_utf8=False,
    pkg_tgt=None,
    pkg_builder=None,
    srcdir=None,
):
    bname = hub.pop_build.init.new(
        name,
        requirements,
        sys_site,
        exclude,
        directory,
        dev_pyinst,
        pyinstaller_runtime_tmpdir,
        build,
        pkg,
        onedir,
        pyenv,
        run,
        locale_utf8,
        pkg_tgt,
        pkg_builder,
        srcdir,
    )
    hub.pop_build.venv.create(bname)
    hub.pop_build.data.version(bname)
    hub.pop_build.build.make(bname)
    hub.pop_build.venv.scan(bname)
    hub.pop_build.venv.mk_adds(bname)
    hub.pop_build.inst.mk_spec(bname)
    hub.pop_build.inst.call(bname)
    if pkg_tgt:
        hub.pop_build.pkg.init.build(bname)
    hub.pop_build.post.report(bname)
    if not no_clean:
        hub.pop_build.post.clean(bname)



# Import python libs
import os


def version(hub, bname):
    """
    Gather the version number of the pop project if possible
    """
    dir_ = hub.pop_build.BUILDS[bname]["dir"]
    name = hub.pop_build.BUILDS[bname]["name"]
    name = name.replace("-", "_")
    path = os.path.join(dir_, name, "version.py")
    _locals = {}
    version = "1"
    try:
        if os.path.isfile(path):
            with open(path) as fp:
                exec(fp.read(), None, _locals)
                version = _locals["version"]
    except Exception:
        pass
    hub.pop_build.BUILDS[bname]["version"] = version



def __init__(hub):
    pass


def build(hub, bname):
    """
    Given the build arguments, Make a package!
    """
    pkg_builder = hub.OPT["pop_build"]["pkg_builder"]
    getattr(hub, f"pop_build.pkg.{pkg_builder}.build")(bname)



# Import python libs
import tempfile
import os
import shutil
import subprocess


def build(hub, bname):
    """
    Build a new package using fpm
    """
    opts = hub.pop_build.BUILDS[bname]
    root = os.path.join(tempfile.mkdtemp(), opts["name"])
    os.makedirs(root)
    # Move files into a tempdir, return config files
    # Set the specific flags, like checksum type, based on the pkg_tgt
    # Run fpm
    config = _prep_tmpdir(
        hub,
        opts["name"],
        root,
        opts["pkg"],
        opts["pkg_tgt"],
        opts["dir"],
        opts["onedir"],
    )
    cmd = _get_cmd(
        opts["name"], root, opts["pkg"], opts["pkg_tgt"], opts["version"], config
    )
    _run_fpm(cmd)
    # shutil.rmtree(root)


def _run_fpm(cmd):
    subprocess.run(cmd)


def _get_fpm_tgt(pkg_tgt):
    """
    Turn the pkg_tgt into a target for fpm
    """
    rpm = (
        "fedora",
        "redhat",
        "rhel",
        "opensuse",
        "suse",
        "cent",
        "centos",
        "sles",
        "sle",
    )
    deb = ("ubuntu", "deb")
    pacman = ("arch", "manjaro")
    if pkg_tgt.lower().startswith(rpm):
        return "rpm"
    elif pkg_tgt.startswith(deb):
        return "deb"
    elif pkg_tgt.lower().startswith(pacman):
        return "pacman"


def _get_cmd(name, root, pkg, pkg_tgt, version, config):
    """
    Return the command line args list to shell out to fpm with
    """
    version = str(pkg.get("version", version))
    fpm_tgt = _get_fpm_tgt(pkg_tgt)
    cmd = ["fpm", "-s", "dir", "-n", name, "-t", fpm_tgt]
    for fn in sorted(list(config)):
        cmd.append("--config-files")
        cmd.append(fn)
    if fpm_tgt == "rpm":
        cmd.append("--rpm-digest")
        cmd.append("sha512")
    cmd.append("--version")
    cmd.append(version)
    cmd.append("-C")
    cmd.append(root)
    return cmd


def _prep_tmpdir(hub, name, root, pkg, pkg_tgt, dir_, onedir):
    """
    Make the tempdir and copy the configured files into it
    """
    dist = os.path.join(dir_, "dist", name)
    configs = set()
    bin_tgt = os.path.join(root, "usr", "bin")
    if onedir:
        dist = os.path.join(dir_, "dist", "run")
        tree_tgt = os.path.join(root, "opt", "run", "bins")
        pbin = os.path.join(f"{os.sep}opt", "run", "bins", "run")
        os.makedirs(os.path.dirname(tree_tgt))
        shutil.copytree(dist, tree_tgt)
        os.makedirs(bin_tgt)
        subprocess.run(["ln", "-sf", pbin, os.path.join(bin_tgt, name)])
    else:
        os.makedirs(bin_tgt)
        shutil.copy(dist, bin_tgt)
    for tpath, spath in pkg.get("config", {}).items():
        tpath = tpath.strip(os.sep)
        src = os.path.join(dir_, spath)
        tgt = os.path.join(root, tpath)
        tgt_dir = os.path.dirname(tgt)
        if not os.path.isdir(tgt_dir):
            os.makedirs(tgt_dir)
        shutil.copy(src, tgt)
        configs.add(tpath)
    for spath in pkg.get("scripts", {}):
        src = os.path.join(dir_, spath)
        if not os.path.isdir(bin_tgt):
            os.makedirs(bin_tgt)
        shutil.copy(src, bin_tgt)
    if pkg_tgt in hub.pop_build.SYSTEMD:
        tgt = os.path.join(root, hub.pop_build.SYSTEMD_DIR)
        for spath in pkg.get("systemd", {}):
            if not os.path.isdir(tgt):
                os.makedirs(tgt)
            src = os.path.join(dir_, spath)
            shutil.copy(src, tgt)
    if pkg_tgt in hub.pop_build.SYSV:
        tgt = os.path.join(root, hub.pop_build.SYSV_DIR)
        for spath in pkg.get("sysv", {}):
            if not os.path.isdir(tgt):
                os.makedirs(tgt)
            src = os.path.join(dir_, spath)
            shutil.copy(src, tgt)
    return configs
